/*
	Copyright (C) 2005 Christopher E. Miller
	
	This software is provided 'as-is', without any express or implied
	warranty.  In no event will the authors be held liable for any damages
	arising from the use of this software.
	
	Permission is granted to anyone to use this software for any purpose,
	including commercial applications, and to alter it and redistribute it
	freely, subject to the following restrictions:
	
	1. The origin of this software must not be misrepresented; you must not
	   claim that you wrote the original software. If you use this software
	   in a product, an acknowledgment in the product documentation would be
	   appreciated but is not required.
	2. Altered source versions must be plainly marked as such, and must not be
	   misrepresented as being the original software.
	3. This notice may not be removed or altered from any source distribution.
*/

private import std.string, std.socket;

private import dfl.all, dfl.winapi;

private import irclib.all, irclib.dflclient;

private import chatinfo;


ChatForm chatForm;
ChatIrc irc;
Font chatFont;
char[] channelName = "#DFL";


void getIrcInfo()
in
{
	assert(irc is null);
}
body
{
	int i;
	ChatInfo f;
	with(f = new ChatInfo)
	{
		showDialog(chatForm);
		if(dialogResult != DialogResult.OK)
		{
			Application.exitThread();
			return;
		}
		
		with(irc = new ChatIrc(tserver))
		{
			if(tnick.length)
				nick = tnick;
			if(tfullName.length)
				fullName = tfullName;
			i = std.string.find(tuser, '@');
			if(i != -1)
				tuser = tuser[0 .. i];
			if(tuser.length)
				userName = tuser;
		}
		
		if(tchan.length)
			channelName = tchan;
		
		// Kill the info form.
		dispose();
	}
	
	// And finally, connect to IRC!
	chatForm.echoInfo("Connecting...");
	irc.connect();
}


void resetIrcInfo()
{
	if(irc)
	{
		chatForm.nicklist.items.clear();
		irc.abort();
		irc = null;
	}
}


void sendAndEchoMessage(char[] target, char[] message)
in
{
	assert(!(irc is null));
}
body
{
	irc.sendMessage(target, message);
	chatForm.echo("<" ~ irc.nick ~ "> " ~ irclib.tools.strip(message));
}


void sendAndEchoAction(char[] target, char[] message)
in
{
	assert(!(irc is null));
}
body
{
	irc.sendAction(target, message);
	chatForm.echo("* " ~ irc.nick ~ " " ~ irclib.tools.strip(message));
}


class ChatIrc: DflIrcClient
{
	this(char[] serverHost, ushort serverPort)
	{
		super(serverHost, serverPort);
	}
	
	
	this(char[] serverHostAndPort)
	{
		super(serverHostAndPort);
	}
	
	
	protected override void onChannelMessageReceived(IrcChannelMessage imsg)
	{
		chatForm.echo("<" ~ imsg.fromNick ~ "> " ~ irclib.tools.strip(imsg.message));
	}
	
	
	protected override void onChannelActionReceived(IrcChannelMessage imsg)
	{
		chatForm.echo("* " ~ imsg.fromNick ~ " " ~ irclib.tools.strip(imsg.message));
	}
	
	
	protected override void onUserNoticeReceived(IrcUserMessage imsg)
	{
		chatForm.echo("-" ~ imsg.fromNick ~ "- " ~ irclib.tools.strip(imsg.message));
	}
	
	
	protected void onTopicChanged(ChannelTopic ctopic)
	{
		chatForm.echoInfo(ctopic.fromNick ~ " has changed the topic to '" ~
			irclib.tools.strip(ctopic.topic) ~ "'");
	}
	
	
	protected void onTopicReply(ChannelTopicReply ctr)
	{
		chatForm.echoInfo("Topic is '" ~ irclib.tools.strip(ctr.topic) ~ "'");
	}
	
	
	protected void onTopicWhoTimeReply(ChannelTopicWhoTimeReply ctwtr)
	{
		chatForm.echoInfo("Topic was set by " ~ ctwtr.setter ~ ".");
	}
	
	
	protected override void onChannelJoin(ChannelJoin cjoin)
	{
		chatForm.addNick(cjoin.fromNick.dup);
		
		if(cjoin.fromNick == nick)
		{
			chatForm.echo("-"); // Spacer.
			chatForm.echoInfo("You have joined the channel.");
		}
		else
		{
			chatForm.echoInfo(cjoin.fromNick ~ " has joined the channel.");
		}
	}
	
	
	private void userleft(IrcFrom who, char[] reason)
	{
		chatForm.removeNick(who.fromNick);
		
		char[] s;
		s = who.fromNick ~ " has left the channel";
		if(reason.length)
			s ~= " (" ~ reason ~ ")";
		else
			s ~= ".";
		chatForm.echoInfo(s);
	}
	
	
	protected override void onChannelPart(ChannelPart cpart)
	{
		userleft(cpart, cpart.reason);
	}
	
	
	protected override void onQuit(IrcQuit iquit)
	{
		if(iquit.fromNick != nick)
			userleft(iquit, iquit.reason);
	}
	
	
	protected override void onChannelKick(ChannelKick ckick)
	{
		chatForm.removeNick(ckick.kickedNick);
		
		chatForm.echoInfo(ckick.fromNick ~ " has kicked " ~ ckick.kickedNick ~
			" from the channel (" ~ irclib.tools.strip(ckick.reason) ~ ")");
	}
	
	
	protected override void onNick(IrcNick inick)
	{
		chatForm.changeNick(inick.fromNick, inick.newNick);
		
		chatForm.echoInfo(inick.fromNick ~ " is now known as " ~ inick.newNick ~ ".");
	}
	
	
	protected override void onLoggedIn()
	{
		sendLine("JOIN " ~ channelName);
	}
	
	
	protected override void onCommand(char[] prefix, char[] cmd, char[] cmdParams)
	{
		char[] s;
		
		switch(cmd)
		{
			case "353": // RPL_NAMREPLY
				if(!_gettingNames)
				{
					_gettingNames = true;
					chatForm.nicklist.items.clear();
				}
				
				ircParam(cmdParams); // My nick.
				ircParam(cmdParams); // Channel type.
				channelName = ircParam(cmdParams);
				s = ircParam(cmdParams); // The names.
				
				foreach(char[] nick; std.string.split(s, " "))
				{
					if(nick.length)
					{
						if(std.string.find(irc.prefixSymbols, nick[0]) != -1)
							nick = nick[1 .. nick.length];
						chatForm.addNick(nick.dup);
					}
				}
				return;
			
			case "366": // RPL_ENDOFNAMES
				_gettingNames = false;
				
				/+
				chatForm.echoInfo("You see here:");
				foreach(Object onick; chatForm.nicklist.items.getEnumerator())
				{
					chatForm.disp.selectedText = " " ~ onick.toString();
				}
				chatForm.disp.selectedText = \r\n;
				+/
				break;
			
			default: ;
		}
		
		super.onCommand(prefix, cmd, cmdParams);
	}
	
	
	/+
	protected override void onLine(char[] line)
	{
		chatForm.echo("[In] " ~ line);
		
		super.onLine(line);
	}
	+/
	
	
	protected override void onHostResolveFail(char[] errmsg)
	{
		chatForm.echoInfo(errmsg);
		
		resetIrcInfo();
		getIrcInfo();
	}
	
	
	protected override void onConnectFail(char[] errmsg)
	{
		chatForm.echoInfo(errmsg);
		
		resetIrcInfo();
		getIrcInfo();
	}
	
	
	protected override void onConnected()
	{
		chatForm.echoInfo("Connected!");
		
		super.onConnected();
	}
	
	
	protected override void onDisconnected()
	{
		chatForm.echoInfo("Disconnected!");
		
		super.onDisconnected();
		
		if(chatForm.windowState == FormWindowState.MINIMIZED)
			chatForm.windowState = FormWindowState.NORMAL;
		
		resetIrcInfo();
		getIrcInfo();
	}
	
	
	private:
	bit _gettingNames = false;
}


class ChatForm: Form
{
	ListBox nicklist;
	TextBox input;
	RichTextBox disp;
	
	
	// Echo a line of text.
	final void echo(char[] s)
	{
		disp.appendText(s ~ \r\n);
	}
	
	
	final void echoInfo(char[] s)
	{
		disp.appendText("\u00A4 " ~ s ~ \r\n);
	}
	
	
	this()
	{
		initializeChatForm();
		
		with(input = new TextBox)
		{
			maxLength = 2000;
			font = chatFont;
			dock = DockStyle.BOTTOM;
			parent = this;
		}
		
		with(nicklist = new ListBox)
		{
			integralHeight = false;
			sorted = true;
			width = 100;
			font = chatFont;
			dock = DockStyle.RIGHT;
			parent = this;
		}
		
		with(new Splitter)
		{
			dock = DockStyle.RIGHT;
			minSize = 60;
			parent = this;
		}
		
		with(disp = new RichTextBox)
		{
			scrollBars = RichTextBoxScrollBars.FORCED_VERTICAL;
			hideSelection = false;
			maxLength = int.max;
			readOnly = true;
			font = chatFont;
			backColor = Color(0xFF, 0xFF, 0xFF);
			foreColor = Color(0);
			detectUrls = false;
			dock = DockStyle.FILL;
			contextMenu = null;
			parent = this;
		}
		
		acceptButton = new Enter;
	}
	
	
	final void addNick(char[] nick)
	{
		if(nicklist.findStringExact(nick) == -1)
			nicklist.items.add(nick);
	}
	
	
	final void removeNick(char[] nick)
	{
		int i;
		i = nicklist.findStringExact(nick);
		if(i != -1)
			nicklist.items.removeAt(i);
	}
	
	
	final void changeNick(char[] oldNick, char[] newNick)
	{
		removeNick(oldNick);
		addNick(newNick);
	}
	
	
	protected override void onVisibleChanged(EventArgs ea)
	{
		super.onVisibleChanged(ea);
		
		if(!visible)
			return;
		
		getIrcInfo();
	}
	
	
	protected override void wndProc(inout Message m)
	{
		const WPARAM SC_RESTORE = 0xF120;
		
		
		switch(m.msg)
		{
			case WM_SYSCOMMAND:
				// There's something wrong with a rich edit when moving the selection
				// in a minimized window, so this should fix it.
				if(m.wParam == SC_RESTORE)
				{
					SendMessageA(disp.handle, WM_VSCROLL, SB_BOTTOM, 0);
				}
				break;
			
			default: ;
		}
		
		super.wndProc(m);
	}
	
	
	private void initializeChatForm()
	{
		// Do not manually edit this block of code.
		//~DFL Designer 0.3 code begins here.
		
		//~DFL MainForm
		text = "DFL IRC";
		clientSize = dfl.drawing.Size(440, 270);
		
		//~DFL Designer 0.3 code ends here.
	}
}


class Enter: IButtonControl
{
	DialogResult dialogResult() // getter
	{
		return DialogResult.OK;
	}
	
	
	void dialogResult(DialogResult result) // setter
	{
	}
	
	
	void notifyDefault(bool byes)
	{
	}
	
	
	void performClick()
	{
		if(irc && irc.isConnected)
		{
			char[] s;
			s = std.string.strip(chatForm.input.text);
			
			if(s.length)
			{
				if(s[0] ==  '/')
				{
					int i;
					char[] cmd, params;
					s = s[1 .. s.length];
					i = std.string.find(s, ' ');
					if(i != -1)
					{
						cmd = s[0 .. i];
						params = s[i + 1 .. s.length];
					}
					else
					{
						cmd = s;
					}
					
					switch(std.string.toupper(cmd))
					{
						case "ME":
							if(!params.length)
								return;
							sendAndEchoAction(channelName, params);
							break;
						
						default:
							chatForm.echo("Unknown command: " ~ cmd);
					}
				}
				else
				{
					sendAndEchoMessage(channelName, s);
				}
			}
			
			chatForm.input.text = null;
		}
	}
}


int main()
{
	int result = 0;
	
	try
	{
		chatFont = new Font("Fixedsys", 10);
		
		Application.run(chatForm = new ChatForm);
	}
	catch(Object o)
	{
		msgBox(o.toString(), "Fatal Error", MsgBoxButtons.OK, MsgBoxIcon.ERROR);
		
		result = 1;
	}
	
	return result;
}

